/*****************************************************************************
    util.c

    Diese Datei ist Teil des Programmpakets 'ACL-tftp'.
    Alle Rechte bei Andreas G. Schindler
                    Klappacher Str. 2i
                    64285 Darmstadt, Deutschland.

    Dieses Programm ist 'Freie Software'; Sie knnen es entsprechend den 
    Bedingungen der 'Allgemeinen ffentlichen GNU-Lizenz' (verffentlicht 
    durch die 'Free Software Foundation') weiter verbreiten und/oder 
    modifizieren. Massgeblich ist die Version 2 oder eine neuere Version
    dieser Lizenz. 

    Dieses Programm wird in der Hoffnung dass es ntzlich is weitergegeben,
    jedoch OHNE JEDE GARANTIE. Dieser Garantieauschluss bezieht sich auch
    auf Eigenschaften wie Marktfhigkeit oder Eignung zu einem bestimmten
    Zweck. Details finden sich in der 'Allgemeinen ffentlichen GNU-Lizenz'.

    Eine Kopie dieser Lizenz sollte in  Programmpaket, dem diese Datei
    entstammt, enthalten sein. Falls nicht, schreiben Sie an die 
    'Free Software Foundation, Inc.', 59 Temple Place, Suite 330, Boston, 
    MA  02111-1307  USA.


    This file is part of the 'ACL-tftp' Package
    Copyright (C) 1999-2003 by Andreas G. Schindler
                               Klappacher Str. 2i
                               Darmstadt, Germany.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*****************************************************************************/
#include "auto.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/timeb.h>
#include <math.h>
#include <ctype.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <errno.h>
#include <limits.h>

#include "globals.h"
#include "util.h"
#include "netio.h"
#include "logfun.h"


#define MODI_UPPER	1
#define MODI_LOWER	2
#define MODI_BS2S	4
#define MODI_NODRIVE	8
#define MODI_FREE	0x4000


#ifndef _GNU_SOURCE
char *strndup(char *src, size_t maxlen)
{
    size_t len;
    char *new;

    if ((len = strlen(src)) > maxlen)
        len = maxlen;

    if ((new = Malloc(len+1)) != NULL)
        memcpy(new, src, len+1);

    return new;
}
#endif /* _GNU_SOURCE */


static char *mapTypes = "IRDBbEeYmdHMSCc";

/*
	print copyright info
*/

void copyright(FILE *fp, int short_form)
{
    if (fp != NULL) {
        fprintf(fp, _("Copyright (C) %s by A.G.Schindler, Darmstadt, "
                      "Germany (andreas@schindlerman.de)\n\n"), 
                COPYRIGHT_YEARS);
        if (!short_form)
            fprintf(fp, 
_("This program is free software; you can redistribute it and/or modify\n"
"it under the terms of the GNU General Public License as published by\n"
"the Free Software Foundation; either version 2 of the License, or\n"
"(at your option) any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful,\n"
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
"GNU General Public License for more details.\n"
"\n"
"You should have received a copy of the GNU General Public License\n"
"along with this program; if not, write to the Free Software\n"
"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n"));
    }
}


/*
	print version info
*/

void version(FILE *fp, char *me)
{
    if (fp != NULL)
        fprintf(fp, _("DARPA tftp protocol %s '%s' V-%d.%2.2d%s%s\n"),
                    me, PACKAGE_NAME, MAJOR_VERSION, MINOR_VERSION,
                    strlen(PATCH_LEVEL) ? " patch level " : "",
                    strlen(PATCH_LEVEL) ? PATCH_LEVEL : "");
}


void logBuild(void)
{
    LOG_MSG2 (LOG_WARNING, 
              _("%s V%d.%d (%s) message level: General:%d, Syslog:%d"),
              PACKAGE_NAME, MAJOR_VERSION, MINOR_VERSION, PACKAGE_DATE,
              msgLevel, maxSyslog);
#ifdef COMPILERVERSION
    LOG_MSG4 (LOG_INFO, _("Compiler: %s"), COMPILERVERSION);
#endif
    LOG_MSG4 (LOG_INFO, _("Compile time options: [%s]"),
#ifdef HAVE_GENDEBUG
    "GENDEBUG "
#endif                          
#ifdef HAVE_MEMDEBUG
    "MEMDEBUG "
#endif                          
#ifdef HAVE_BACKTRACE
    "BACKTRACE "
#endif                          
#ifdef HAVE_BLOCKSIZE2
    "BLOCKSIZE2 "
#endif                          
#ifdef FILE_CACHE
    "FILECACHE "
#endif
    "");                          
}

void showBuild(FILE *fp)
{
    fprintf(fp, _("%s V%d.%d (%s) message level: General:%d, Syslog:%d"),
            PACKAGE_NAME, MAJOR_VERSION, MINOR_VERSION, PACKAGE_DATE,
            msgLevel, maxSyslog);
#ifdef COMPILERVERSION
    fprintf(fp, _("Compiler: %s"), COMPILERVERSION);
#endif
    fprintf(fp, _("Compile time options: [%s]"),
#ifdef HAVE_GENDEBUG
    "GENDEBUG "
#endif                          
#ifdef HAVE_MEMDEBUG
    "MEMDEBUG "
#endif                          
#ifdef HAVE_BACKTRACE
    "BACKTRACE "
#endif                          
#ifdef HAVE_BLOCKSIZE2
    "BLOCKSIZE2 "
#endif                          
#ifdef FILE_CACHE
    "FILECACHE "
#endif
    "");                          
}

static char *editMap(char *in, int modifier)
{
    char *out, *po;

    if (in == NULL)
        return NULL;

    if (modifier & MODI_NODRIVE) {
        po = strchr(in, ':');
        if (po == NULL)
            po = in;
        else
            po++;
    }
    else
        po = in;

    out = Strdup(po);

    if (out == NULL) {
        errNoDynmem;
    }
    else {
        for (po = out; *po != '\0'; po++) {
            if ((modifier & MODI_UPPER) && islower(*po)) 
                *po = toupper(*po);
            if ((modifier & MODI_LOWER) && isupper(*po)) 
                *po = tolower(*po);
            if ((modifier & MODI_BS2S) && (*po == '\\')) 
                *po = '/';
        }
    }

    if (modifier & MODI_FREE)
        Free(in);

    return out;
}


char *baseName (char *path)
{
    char *slash;

    if (path == NULL)
        return NULL;

    if ((slash = strrchr(path, '/')) != NULL)
        return slash+1;
 
    return path;
}


char *dirName (char *path)
{
    char *name, *ctx, *slash;

    if (path == NULL)
        return NULL;

    if ((name = zstrtok(path, "\t ", &ctx)) == NULL)
        return NULL;

    slash = strrchr(name, '/');

    if (slash == NULL)
        *name = '\0';
    else
        *slash = '\0';

    return name;
}

/*
	check if we're member of a specific group
	checks effectife group and secondary groups
*/

int isGroupMember (gid_t gid)
{
#define GLISTSIZE	32
    gid_t glist[GLISTSIZE], mygid;
    int groups, i;

    if (gid == (mygid =getegid()))
        return 1;

    if ((groups = getgroups(GLISTSIZE, glist)) > 0) 
        for (i=0; i<groups; i++)
            if (glist[i] == mygid)
                return 1;
    
    return 0;
#undef GLISTSIZE
}

/*
	checks the accessability of a file in resepct to the
	desired operation (ACC_READ, ACC_WRITE, ACC_CREATE)
*/

int fileAccess (char *file, int check) 
{
    int ios, sos;
    char *dir = NULL;
    struct stat sb;

    ios = stat(file, &sb);

    switch (check) {
        case ACC_READ:
            if (ios)
                return errno;
            if ((S_IROTH & sb.st_mode))
                return 0;
            if ((geteuid() == sb.st_uid) && (S_IRUSR & sb.st_mode))
                return 0;
            if (isGroupMember(sb.st_gid) && (S_IRGRP & sb.st_mode))
                return 0;
            break;
        case ACC_WRITE:
            if (ios)
                return errno;
            if ((S_IWOTH & sb.st_mode))
                return 0;
            if ((geteuid() == sb.st_uid) && (S_IWUSR & sb.st_mode))
                return 0;
            if (isGroupMember(sb.st_gid) && (S_IWGRP & sb.st_mode))
                return 0;
            break;
        case ACC_CREATE:
            if (!ios)
                return EEXIST;
            if ((dir = dirName(file)) == NULL)
                return ENOMEM;
            sos = stat(dir, &sb);
            Free(dir);
            if (sos)
                return errno;
            if ((S_IRWXO & sb.st_mode) == S_IRWXO)
                return 0;
            if ((geteuid() == sb.st_uid) && ((S_IRWXU & sb.st_mode) == S_IRWXU))
                return 0;
            if (isGroupMember(sb.st_gid) && ((S_IRWXG & sb.st_mode) == S_IRWXG))
                return 0;
            break;
        case ACC_EXEC:
            if (ios)
                return errno;
            if ((S_IXOTH & sb.st_mode))
                return 0;
            if ((geteuid() == sb.st_uid) && (S_IXUSR & sb.st_mode))
                return 0;
            if (isGroupMember(sb.st_gid) && (S_IXGRP & sb.st_mode))
                return 0;
            break;
        default:
            return EINVAL;
    }
    return EACCES;
}

/*
	checks exsitence & accessability of a directory
	fails, if the target is not a directory (soft links transparent !)
	accessability constants (R_OK, W_OK, X_OK) defined in access(2)

	all return codes according to errno(3) 
*/

int dirAccess (char *dirname, int check)
{
    int myuser, mygroup;
    struct stat sb;

    if (stat(dirname, &sb))
        return errno;

    if (!S_ISDIR(sb.st_mode))
        return ENOTDIR;

    myuser  = (geteuid() == sb.st_uid);
    mygroup = isGroupMember(sb.st_gid);

    if (check & R_OK)
        if (!(S_IROTH & sb.st_mode))
            if (!myuser || !(S_IRUSR & sb.st_mode))
                if (!mygroup || !(S_IRGRP & sb.st_mode))
                    return EACCES;
    
    if (check & W_OK)
        if (!(S_IWOTH & sb.st_mode))
            if (!myuser || !(S_IWUSR & sb.st_mode))
                if (!mygroup || !(S_IWGRP & sb.st_mode))
                    return EACCES;

    if (check & X_OK)
        if (!(S_IXOTH & sb.st_mode))
            if (!myuser || !(S_IXUSR & sb.st_mode))
                if (!mygroup || !(S_IXGRP & sb.st_mode))
                    return EACCES;
    return 0;
}


/*
	get stat of parent directory of 'filename'
*/

int dpstat (char *filename, struct stat *sb)
{
    char *dirname;
    int  ios;

    if ((dirname = dirName(filename)) == NULL)
        return ENOMEM;

    memset(sb, 0, sizeof(struct stat));

    if (strlen(dirname))
        ios = stat(dirname, sb);
    else
        ios = stat(".", sb);

    Free(dirname);

    if (ios < 0)
        return errno;

    if (!S_ISDIR(sb->st_mode))
        return ENOTDIR;

    return 0;
}


/*
	remove nonprintable characters from end of a string
*/

void trim (char *txt)
{
    char *p;

    for (p = txt+strlen(txt)-1; p >= txt; p--)
        if (!isgraph((int)*p))
            *p = '\0';
}


/*
        Multi-thread "strtok" (identical behaviour) but no static
        storage. All context necessary to deal with athread is stored in
        the "context" pointer
*/
char *xstrtok (char *buf, char *pat, char **context)
{
    char *ptr, *result;

    if (context  == NULL) return (NULL);
    if (*pat == '\0') return (*context=NULL);
    if (buf != NULL) *context = buf;
    if ((result = *context) == NULL) return (NULL);

    for (; *result != '\0'; result++)
        if (strchr(pat,(int)*result) == NULL) break;
    if (*result == '\0') return (*context=NULL);

    ptr = strpbrk(result, pat);

    if (ptr != NULL) *ptr++ = '\0';
    *context = ptr;

    return(result);
}
/*
	Multi-thread "strtok" (identical behaviour) but no static
	storage. All context necessary to deal with a thread is stored in
	the "context" pointer.

	additionally, there is a quoting parameter list, that defines
	quoting character pairs. Text between one of these pairs is
	excepted from the separator matching.
*/
char *qstrtok (char *buf, char *pat, char **context, char *quotes)
{
    char *ptr = NULL, *result;

    if (context  == NULL) return (NULL);
    if (*pat == '\0') return (*context=NULL);
    if (buf != NULL) *context = buf;
    if ((result = *context) == NULL) return (NULL);

    for (; *result != '\0'; result++)
	if (strchr(pat,(int)*result) == NULL) break;

    if (*result == '\0') return (*context=NULL);

    for (; *quotes != '\0'; quotes++)
	if (*result == *quotes) {
            if (*(++quotes) != '\0') {
                if (*(++result) == '\0') return (*context=NULL);
                    ptr = strchr(result, (int)(*quotes));
            }
            break;
        }

    if (*quotes == '\0')
        ptr = strpbrk(result, pat);

    if (ptr != NULL) *ptr++ = '\0';
    *context = ptr;

    return(result);
}

/*
	Basically the same a "xstrtok", but doesn't modify the contents of
	the input buffer. Instead each aquired token is copied to dynamically
	allocated buffers (by using "malloc"). The caller is responsible
	for freeing these buffers if needed.
*/
char *zstrtok (char *buf, char *pat, char **context)
{
    char *ptr, *temp, *result;
    int len;

    if (context  == NULL) return (NULL);
    if (*pat == '\0') return (*context=NULL);
    if (buf != NULL) *context = buf;
    if ((temp = *context) == NULL) return (NULL);

    for (; *temp != '\0'; temp++)
	if (strchr(pat,(int)*temp) == NULL) break;
    if (*temp == '\0') return (*context=NULL);

    ptr = strpbrk(temp, pat);

    if (ptr == NULL) {
	len = strlen(temp);
	*context = NULL;
    }
    else {
	len = (int)(ptr - temp);
	*context = (ptr+1);
    }

    result = (char *)Malloc(len+1);

    if (result == NULL) {
        errNoDynmem;
	*context = NULL;
    }
    else {
	memcpy(result,temp,len);
	*(result+len) = '\0';
    }

    return(result);
}

/*
	Basically the same a "zstrtok", but operates on a fixed length
	buffer. The separator is always NUL.
	'context' and 'remlen' are auxiliary storage used to provide reentancy.
	They must not be changed by the caller.
	'reslen' is an optional parameter to receive the lenght of the
	returned string.
*/
char *zstrtok0 (char *buf, size_t buflen, char **context, size_t *remlen, 
                size_t *reslen)
{
    char *temp, *result;

    if ((context == NULL) || (remlen  == NULL)) return (NULL);
    if (buf != NULL) {
        *context = buf;
        *remlen  = buflen;
    }
    if ((temp = *context) == NULL) return (NULL);
    if (*remlen == 0) return (NULL);

    for (; remlen; *remlen -= 1) {
        if (*temp++ == '\0') {
            if (*remlen)
                *remlen -= 1; 
            break;
        }
    }
    
    result = strndup(*context, temp - *context);

    if (result == NULL)
	*context = NULL;
    else {
        *context = temp;
        if (reslen != NULL)
            *reslen = temp - *context;
    }

    return(result);        
}


/*
        check file map string for illegal placeholder codes
        on error the position of the character inquestion is returned
*/

int checkFileMap (char *map)
{
    char *m;
    for (m = map; *m != '\0'; m++)
        if (*m == '%')
            if (strchr(mapTypes, *(m+1)) == NULL)
                return ((int)(m-map));
    return 0;
}


char *xlateFileMap (char *path, char *map, struct sockaddr_in *client,
                    char *realpath, int dir)
{
#define MINLEN 64

    char *new = NULL, *mem, *start = NULL, *next = NULL, buffer[8], *dot;
    int  size = 0, rest = 0, len, done, modifier = 0;
    time_t now;
    struct tm lt;

    now = time(NULL);
    memcpy(&lt, localtime(&now), sizeof(struct tm));

    for (; *map != '\0'; map++) {
        mem = NULL;
        if ((*map == '%') && (*(map+1) != '\0')) {
            len = 0;
            for (done = 0; !done;) {
                switch (*(map+1)) {
                    case '+':		/* all upper case 	*/
                        modifier |= MODI_UPPER;
                        map++;
                        break;
                    case '-':		/* all lower case 	*/
                        modifier |= MODI_LOWER;
                        map++;
                        break;
                    case '=':		/* backslash to slash	*/
                        modifier |= MODI_BS2S;
                        map++;
                        break;
                    case ':':		/* no drive letter	*/
                        modifier |= MODI_NODRIVE;
                        map++;
                        break;
                    default:
                        done = 1;
                        break;
                }
            }
            switch (*(++map)) {
                case 'I':		
                    if (dir == RT_RRQ)
                        start = "r";
                    if (dir == RT_WRQ)
                        start = "w";
                    else
                        start = NULL;
                    if (start != NULL)
                        len = strlen(start);
                    break;

                case 'R':
                    start = editMap(realpath, modifier);
                    if (start != NULL)
                        len = strlen(start);
                    break;

                case 'D':
                    mem = start = editMap(dirName(path), modifier|MODI_FREE);
                    if (start != NULL)
                        len = strlen(start);
                    break;

                case 'B':
                    start = editMap(baseName(path), modifier);
                    if (start != NULL)
                        len = strlen(start);
                    break;

                case 'b':
                    start = editMap(baseName(path), modifier);
                    if (start != NULL) {
                        if ((dot = strchr(start,'.')) != NULL)
                            *dot = '\0';
                        len = strlen(start);
                    }
                    break;

                case 'E':
                case 'e':
                    start = editMap(baseName(path), modifier);
                    if (start != NULL) {
                        if ((start = strchr(start,'.')) != NULL)
                            len = strlen(++start);
                    }
                    break;

                case 'Y':
                    snprintf(buffer, sizeof(buffer), "%4.4d", lt.tm_year+1900);
                    start = buffer;
                    len = 4;
                    break;

                case 'm':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_mon+1);
                    start = buffer;
                    len = 2;
                    break;

                case 'd':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_mday);
                    start = buffer;
                    len = 2;
                    break;

                case 'H':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_hour);
                    start = buffer;
                    len = 2;
                    break;

                case 'M':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_min);
                    start = buffer;
                    len = 2;
                    break;

                case 'S':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_sec);
                    start = buffer;
                    len = 2;
                    break;

                case 'C':
                case 'c':
                    if (client == NULL) {
                        len = 0;
                    }
                    else {
                        struct hostent *hp;
                        if (doResolv & RES_CONFIG)
                            hp = gethostbyaddr((char *)&client->sin_addr.s_addr,
                                               sizeof(in_addr_t),AF_INET);
                        else
                            hp = NULL;

                        if ((hp == NULL) || (hp->h_name == NULL)) {
                            start = inet_ntoa(client->sin_addr);
                        }
                        else {
                            start = hp->h_name;
                            if (*map == 'c') {
                                char *dot;
                                if ((dot = strrchr(hp->h_name, '.')) != NULL)
                                    start = dot+1;
                            }
                        }
                        len = strlen(start);
                        break;
                    }
/*
	Remove unknown placeholders to avoid problems with 'printf'-family
	style functions used by the logging facility
*/
                default:
                    len = 0;
                    break;
            }
        }
        else {
            start = map;
            len = 1;
        }
        if (len > 0) {
            while (rest <= len) {
                int olds;
                olds = size;
                size += ((len/MINLEN)+1)*MINLEN;
                if ((new = (char *) Realloc(new, size)) == NULL) {
                    errNoDynmem;
                    return NULL;
                }
                rest += (size - olds);
                next = new + size - rest;
            }
            memcpy(next, start, len);
            rest -= len;
            next += len;
        }
        if (mem != NULL)
            Free(mem);
    }
    if (new == NULL) {
        new = next = (char *)Malloc(1);
    }
    if (new == NULL)
        errNoDynmem;
    else
        if (next != NULL)
            *next = '\0';
    return new;
#undef MINLEN
}    


char *xlatePath (char *path)
{
    char *newpath = NULL, *pnew, *start = NULL, buffer[32];
    int  rest = 0, len, newlen = 0;
    time_t now;
    struct tm lt;

    now = time(NULL);
    memcpy(&lt, localtime(&now), sizeof(struct tm));

    for (; *path != '\0'; path++) {
        if ((*path == '%') && (*(path+1) != '\0')) {
            switch (*(++path)) {
                case 'Y':
                    snprintf(buffer, sizeof(buffer), "%4.4d", lt.tm_year+1900);
                    start = buffer;
                    len = 4;
                    break;

                case 'm':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_mon+1);
                    start = buffer;
                    len = 2;
                    break;

                case 'd':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_mday);
                    start = buffer;
                    len = 2;
                    break;

                case 'H':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_hour);
                    start = buffer;
                    len = 2;
                    break;

                case 'M':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_min);
                    start = buffer;
                    len = 2;
                    break;

                case 'S':
                    snprintf(buffer, sizeof(buffer), "%2.2d", lt.tm_sec);
                    start = buffer;
                    len = 2;
                    break;

                case 'p':
                    snprintf(buffer, sizeof(buffer), "%d", getpid());
                    start = buffer;
                    len = strlen(buffer);
                    break;
/*
	Remove unknown placeholders to avoid problems with 'printf'-family
	style functions used by the logging facility
*/
                default:
                    continue;
            }
        }
        else {
            start = path;
            len = 1;
        }
        if (len > 0) {
            if (rest < len) {
                if (newlen == 0) {
                    newlen = rest = strlen(path);
                }
                else {
                    newlen += len;
                    rest   += len;
                }
                if ((newpath = (char *) Realloc(newpath, newlen+1)) == NULL) {
                    errNoDynmem;
                    return NULL;
                }
                pnew = newpath + newlen - rest;
            }
            rest -= len;
            while (len--)
                *pnew++ = *start++;
        }
    }
    *pnew = '\0';
    return newpath;
}    


char *fmtstat (struct stat *st)
{
    static char fmtbuf[64];
    struct passwd *pp;
    struct group *gp;
    char *pw_name, *gr_name;

    if((gp = getgrgid(st->st_gid)) == NULL)
      gr_name = "?";
    else
      gr_name = gp->gr_name;

    if((pp = getpwuid(st->st_uid)) == NULL)
      pw_name = "?";
    else
      pw_name = pp->pw_name;

    snprintf (fmtbuf, sizeof(fmtbuf), 
              "%3.3o %s:%s", st->st_mode & 0x1ff, pw_name, gr_name);

    return fmtbuf;
}

/*
	check netmask for CIDR conformity

	on success returns number of bits in netmask (0-32)
	or -1 on error
*/

int iscidr (in_addr_t mask)
{
    in_addr_t bit = 0x80000000;
    int i, n;

    for (i=32; (i) && (mask & bit); i--, bit >>= 1);

    for (n=i ; n ; n--, bit >>= 1)
        if ((mask & bit))
            return -1;

    return 32-i;
} 

/*
	check network address in CIDR notation for conformity

	valid formats:	1.) dotted decimal:	a.b.c.d/e
	                2.) hex:		xxxxxxxxx/e

	2nd argument, if != 0 causes truncation of address before
	CIDR mask (success only);

	on success returns number of bits in netmask, else -1
*/

char *cidrsplit (char* addr)
{
    char *sp;
    uint mbits;

    if ((sp = strrchr(addr, '/')) == NULL) 
        return (C_IPMASK_HOST);

    if (sscanf((sp+1), "%u", &mbits) != 1)
        return NULL;

    *sp++ = '\0';

    return (ccidr2mask((int)mbits));
}

/*
	convert CIDR network bitfield number to traditional netmask
	in 4-octet binary format:	27 -> 255.255.255.224
	-	REENTRANT	-
*/	
in_addr_t cidr2mask(int cidrbits)
{
    in_addr_t mask = 0xffffffff;

    if (cidrbits <  0) cidrbits =  0;
    if (cidrbits > 32) cidrbits = 32;
    
    return (in_addr_t)((mask << (32-cidrbits)));
}

/*
	convert CIDR network bitfield number to traditional netmask
	in 4-octet ascii format:	27 -> "255.255.255.224"
	-	NON REENTRANT	-
*/	
char *ccidr2mask(int cidrbits)
{
    struct in_addr sia;

    sia.s_addr = htonl(cidr2mask(cidrbits));

    return (inet_ntoa(sia));
}

/*
	convert signed integer to ascii
	-	REENTRANT	-
*/
int i2ar (int i, char *a, size_t alen)
{
    return ((snprintf(a,alen,"%d",i) == 1) ? 0 : 1);
}

/*
	convert signed integer to ascii
	-	NON REENTRANT	-
*/
char *i2a (int i)
{
    static char a[32];

    i2ar (i, a, sizeof(a));
    return a;
}

int stricmp(char *s1, char *s2)
{
#ifdef HAVE_STRCASECMP
    return strcasecmp(s1, s2);
#else
    char c1 , c2;

    for (; *s1 != '\0'; s1++) {
        c1 = (islower(*s1)) ? *s1 : tolower(*s1);
        for (; *s2 != '\0'; s2++) {
            c2 = (islower(*s2)) ? *s2 : tolower(*s2);
            if (c1 != c2)
                return 1;
        }
    }
    return 0;
#endif
}

int strnicmp(char *s1, char *s2, size_t len)
{
#ifdef HAVE_STRNCASECMP
    return strncasecmp(s1, s2, len);
#else
    char c1 , c2;

    for (; (*s1 != '\0') && len; s1++, len--) {
        c1 = (islower(*s1)) ? *s1 : tolower(*s1);
        for (; *s2 != '\0'; s2++) {
            c2 = (islower(*s2)) ? *s2 : tolower(*s2);
            if (c1 != c2)
                return 1;
        }
    }
    return 0;
#endif
}

int cutLine (char *line, int *argc, char **argv, int maxargs, char **ctx)
{
    static char *sep = " \t", *quo = "''''", *tok;
    int i=0;

    for (tok = qstrtok(line, sep, ctx, quo); tok != NULL; 
         tok = qstrtok(NULL, sep, ctx, quo)) {
        if (i >= maxargs)
            return -1;
        argv[i++] = tok;
    }
    *argc = i;
    return i;

}

static ulong tMask (int bitNo)
{
    ulong one = 1;

    return (one << (bitNo & ((sizeof(one)*8)-1)));
}

void lbitSet (ulong *mask, int bitNo)
{
    *mask |= tMask(bitNo);
}

void lbitClr (ulong *mask, int bitNo)
{
    *mask &= ~(tMask(bitNo));
}

int lbitTst (ulong *mask, int bitNo)
{
    return (*mask & tMask(bitNo));
}

int lbitTstSet (ulong *mask, int bitNo)
{
    ulong temp;

    temp = *mask;
    *mask |= tMask(bitNo);
    return (temp & tMask(bitNo));
}

int lbitTstClr (ulong *mask, int bitNo)
{
    ulong temp;

    temp = *mask;
    *mask &= ~(tMask(bitNo));
    return (temp & tMask(bitNo));
}

char *dynMemStat(void)
{
#ifdef HAVE_STRUCT_MSTATS
    static char buffer[256];
    struct mstats mstats;

    mstats = mstats();

    snprintf (buffer, sizeof(buffer), "D/Mem By u:%u f:%u, Ch u:%u f:%u\n",
              mstats.bytes_used,mstats.bytes_free,
              mstats.chunks_used,mstats.chunks_free);
    return buffer;
#else
    return NULL;
#endif
}

#ifdef HAVE_GETTIMEOFDAY
struct timeval uftime(struct timeval *from)
{
    struct timeval tv;
    struct timezone tz;

    (void)gettimeofday(&tv, &tz);

    if (from != NULL)
        tv = tvDiff (tv, *from);

    return tv;
}
#else
struct timeval uftime(struct timeval *from)
{
    struct timeb tb;
    struct timeval tv;

    (void)ftime(&tb);

    tv.tv_sec = tb.time;
    tv.tv_usec = 1000 * (long)tb.millitm;

    if (from != NULL)
        tv = tvDiff (tv, *from);

    return tv;
}
#endif /* HAVE_GETTIMEOFDAY */

int asc2tv(char *numstr, struct timeval *tvp)
{
    float rnum;

    if (sscanf(numstr,"%g",&rnum) != 1)
        return 1;

    *tvp = double2tv((double)rnum);

    return 0;
}

/*
	Exponential backoff support routines

	initBackoff:	Initialize exponential backup. 
			Intended to save runtime CPU load.
			Currently there is no use of 'strength', which
			shouls be used to contol the 'agressivity'
			of the exponential backoff algorithm

	proBackoff:	Computes the progressive backoff time
			If the base timeout is positive, there is no
			progressive backoff. The returned value is
			the base timeout regardless of the retry count.
			If the base timout is negative, a poor man's
			kind of progressive backoff is provided in the
			following way:
			- The overall timeout, i.e. base-timeout * max-retry
			  is kept constant
			- The individual timeous are increasing with the
			  current retry count.
			The 'strength' value kept in tout_t:xbfac is currently
			unused. So, the agressivity of the backoff algorithm
			is constant and rather weak.

	listBackoff:	Dumps a list of all possible timout values in
			respect to the retry count (Debug).
*/


int initBackoff(tout_t *top, double strength)
{
    top->xbfac = strength;
    return 0;
}


struct timeval proBackoff(tout_t *top, int selector, int curRetry)
{
    struct timeval *tvp;
    double tmo;

    switch (selector) {
        case EBS_SEND:
            tvp = &top->send;
            break;
        case EBS_SACK:
            tvp = &top->sack;
            break;
        case EBS_RECV:
            tvp = &top->recv;
            break;
        case EBS_RACK:
            tvp = &top->rack;
            break;
        case EBS_CONN:
            tvp = &top->conn;
            break;
        case EBS_IDLE:
        default:
            tvp = &top->idle;
            break;
    }

    if ((tvp->tv_sec >= 0) && (tvp->tv_usec >= 0))
        return *tvp; 

    tmo = (double)(curRetry+1) * -2.0 * tv2double(tvp) / (double)(top->retry+1);

    return (double2tv(tmo));
}


char *listBackoff(tout_t *top, int selector)
{
    int i, rest, l;
    char temp[64], *p;
    static char buf[256];
    struct timeval tmo;

    for (i=0, p=buf, rest=sizeof(buf)-3 ; i<top->retry; i++) {
        tmo = proBackoff(top, selector, i);
        snprintf(temp, sizeof(temp), "%d:%.1f ", i, tv2double(&tmo));
        l = strlen(temp);
        if (rest > l) {
            strcpy(p, temp);
            rest -= l;
            p += l;
        }
        else {
            strcpy(p, "...");
            break;
        }
    }
    return buf;
}
        

double tv2double(struct timeval *tvp)
{
    return ((double)tvp->tv_sec + ((double)tvp->tv_usec/1000000.0));
}


struct timeval double2tv(double d)
{
    struct timeval tv;
    double rem, ipa;

    rem        = modf(d, &ipa);
    tv.tv_sec  = (long)ipa;
    tv.tv_usec = (long)(rem *1000000.0);

    return tv;
}

struct timeval int2tv(long ip, long fp)
{
    struct timeval tv;

    tv.tv_sec  = ip;
    tv.tv_usec = fp;
    return tv;
}

char *tv2asc(struct timeval *tv)
{
    static char buf[64];
    snprintf(buf, sizeof(buf), "%.1f", tv2double(tv));
    return buf;
}

char *timer2asc (tout_t *tp, int what)
{
    static char buf[256];
    int buflen = 0;

    buf[0] = '\0';

    if (what & T2A_XFER) {
        buflen = strlen(buf);
        snprintf(buf+buflen, sizeof(buf)-buflen,
                 "Se:%.1f SA:%.1f Re:%.1f RA:%.1f ",
                 tv2double(&tp->send), tv2double(&tp->sack),
                 tv2double(&tp->recv), tv2double(&tp->rack));
    }
    if (what & T2A_SETUP) {
        buflen = strlen(buf);
        snprintf(buf+buflen, sizeof(buf)-buflen, "Id:%.1f Co:%.1f ",
                 tv2double(&tp->idle), tv2double(&tp->conn));
    }
    if (what & T2A_RET) {
        buflen = strlen(buf);
        snprintf(buf+buflen, sizeof(buf)-buflen, "Rt:%d ", tp->retry);
    }
    if (what & T2A_SPEC) {
        int eflag = tp->eflag & (SD_PERMIT|SD_DENY|SD_KEEP|SD_UNSET);
        if (eflag) {
            buflen = strlen(buf);
            snprintf(buf+buflen, sizeof(buf)-buflen, "Ef:%d ", eflag);
        }
        if (tp->eflag & SD_OLDPROTO) {
            buflen = strlen(buf);
            snprintf(buf+buflen, sizeof(buf)-buflen, "Op ");
        }
    }
    buflen = strlen(buf);
    if (buf[buflen-1] == ' ') 
        buf[buflen-1] = '\0'; 
    return buf;
}

struct timeval tvDiff (struct timeval t1, struct timeval t2)
{
    struct timeval diff;

    diff.tv_sec  = t1.tv_sec  - t2.tv_sec;
    diff.tv_usec = t1.tv_usec - t2.tv_usec;

    while (diff.tv_usec < 0) {
        diff.tv_sec  -= 1;
        diff.tv_usec += 1000000;
    } 

    if (diff.tv_sec < 0)
        diff.tv_sec  = diff.tv_usec = 0;

    return diff;
}


struct timeval tvSum (struct timeval t1, struct timeval t2)
{
    struct timeval sum;

    sum.tv_usec = t1.tv_usec + t2.tv_usec;
    sum.tv_sec  = t1.tv_sec  + t2.tv_sec;

    while (sum.tv_usec >= 1000000) {
        sum.tv_sec  += 1;
        sum.tv_usec -= 1000000;
    }

    return sum;
}


struct timeval tvDivide (struct timeval t1, int fac)
{
    struct timeval res;
    
    res.tv_sec   = t1.tv_sec  / fac;
    res.tv_usec  = t1.tv_usec / fac;
    res.tv_usec += t1.tv_sec  % fac;

    return res;
}

int tvCmpGT (struct timeval t1, struct timeval t2)
{
    if (t1.tv_sec > t2.tv_sec)
        return 1;
    else if (t1.tv_sec < t2.tv_sec)
        return 0;
    return (t1.tv_usec > t2.tv_usec);
}

int tvCmpGE (struct timeval t1, struct timeval t2)
{
    if (t1.tv_sec > t2.tv_sec)
        return 1;
    else if (t1.tv_sec < t2.tv_sec)
        return 0;
    return (t1.tv_usec >= t2.tv_usec);
}

int parseTimes(char *arg, tout_t *tp)
{
    int k, ival=0;
    static char *tokList[] = { "se","sa","re","ra","co","id","rt","ef","op" };
    static char *sep = ":=";
    char *ctx, *tok, *val;
    struct timeval tval;

    if ((arg == NULL) || (strlen(arg) == 0))
        return 0;

    tok = xstrtok(arg,  sep, &ctx);
    val = xstrtok(NULL, sep, &ctx);

    if ((tok == NULL) || (val == NULL))
        return -1;

    for (k=((sizeof(tokList)/sizeof(tokList[0]))-1); k>=0; k--) {
        if (!strncasecmp(tok, tokList[k],2))
            break;
    }
    if (k < 0)
        return -2;
       
    switch (k) {
      case 5:
        if (asc2tv(val, &tval))
            return -3;
        break;
      case 0:
      case 1:    
      case 2:    
      case 3:    
      case 4:    
        if (asc2tv(val, &tval))
            return -3;
        if (!tval.tv_sec && !tval.tv_usec)
            return -4;
        break;
      case 6:
        if (sscanf(val,"%d",&ival) != 1)
            return -3;
        if (ival == 0)
            return -4;
        break;
      case 7:
        if (sscanf(val,"%d",&ival) != 1)
            return -3;
        if ((ival != SD_PERMIT) && (ival != SD_DENY))
            return -4;
        break;
      case 8:
        if (sscanf(val,"%d",&ival) != 1)
            return -3;
        break;
    }

    switch (k) {
      case 0:
        tp->send = tval;
        break;
      case 1:    
        tp->sack = tval;
        break;
      case 2:    
        tp->recv = tval;
        break;
      case 3:    
        tp->rack = tval;
        break;
      case 4:    
        tp->idle = tval;
        break;
      case 5:    
        tp->conn = tval;
        break;
      case 6:
        tp->retry = ival;
        break;
      case 7:
        tp->eflag = (tp->eflag & (~SD_UNSET)) | (ival ? SD_PERMIT : SD_DENY);
        break;
      case 8:
        if (ival)
            tp->eflag |= SD_OLDPROTO;
        break;
    }  
    return 1;
}

int parseCTimes (char *arglist, tout_t *tp)
{
    char *ctx, *tok;
    int rtc, sts=0;

    if (arglist == NULL)
        return 0;

    for (tok=xstrtok(arglist,",",&ctx); tok!=NULL; tok=xstrtok(NULL,",",&ctx)){
        if ((rtc = parseTimes(tok, tp)) < 0)
            return -1;
        sts += rtc;
    }

    return sts;
}

void startTimer(V_L struct timeval *timer)
{
    *timer = uftime(NULL);
}


int checkTimer(struct timeval timer, struct timeval compare_to)
{
    struct timeval itimer = timer;
    return tvCmpGE(uftime(&itimer), compare_to);
}


char *strdupNoNull(char *string, int len, char repl)
{
    int i;
    char *new, *np;

    if ((new = np = Malloc(len+1)) == NULL)
        return NULL;

    for (i=0; i<len; i++)
        if (string[i] == '\0')
            *np++ = repl;
        else
            *np++ = string[i];

    *np = '\0';
    return new;
}

char *mflag2name(int flag)
{
    switch (flag) {
        case MF_NONE:
            return "NONE";
            break;
        case MF_PRESET:
            return "PRESET";
            break;
        case MF_DEFAULT:
            return "DEFAULT";
            break;
        case MF_CMD:
            return "CMD";
            break;
        case MF_CONFIG:
            return "CONFIG";
            break;
        case MF_RUN:
            return "RUN";
            break;
        default:
            return "";
    }
}

char *mflag2txt (int flag)
{
    static char buf[64];
    static int flaglist[] = 
        {MF_NONE,MF_PRESET,MF_DEFAULT,MF_CMD,MF_CONFIG,MF_RUN};
    int i, j;

    for (i=j=0, buf[0]=0; i<(int)(sizeof(flaglist)/sizeof(flaglist[0])); i++) {
        if ((flag & flaglist[i])) {
            if (j++) strcat(buf,"|");
            strncat(buf, mflag2name(flaglist[i]), sizeof(buf));
        }
    }
    return buf;
}

char shortMF (int flag)
{
    struct flaglist { int flag; char code; };
    static struct flaglist flaglist[] = 
        {{MF_NONE,   'N'},
         {MF_PRESET, 'P'},
         {MF_DEFAULT,'D'},
         {MF_CMD,    'C'},
         {MF_CONFIG, 'F'},
         {MF_RUN,    'R'}};
    int i;

    for (i=0; i<(int)(sizeof(flaglist)/sizeof(flaglist[0])); i++)
        if (flag & flaglist[i].flag)
            return flaglist[i].code;
    return '?';
}

int globCheck (const char *func, int line, int *Mflag, char *name, int flag)
{
    char *fmt;

    fmt = _("%s@%d: mod global[%s] by[%s] flag[%s]: %s");

    if ((*Mflag & MF_CMD) && !(flag & MF_CMD)) {
        LOG_MSG6 (LOG_DEBUG, fmt, func, line, name, 
                  mflag2name(flag), mflag2txt(*Mflag), "*denied*");
        return 1;
    }
    else {
        LOG_MSG6 (LOG_DEBUG, fmt, func, line, name, 
                  mflag2name(flag), mflag2txt(*Mflag), "*ok*");
        *Mflag = flag;
        return 0;
    }
}    

int txt2uid (char *text, uid_t *uid)
{
    struct passwd *sp;
    int iuid;

    if ((sp = getpwnam(text)) == NULL) {
        if (sscanf(text,"%d",&iuid) == 1)
            sp = getpwuid((uid_t)iuid);
    }
    if (sp == NULL) {
        LOG_MSG1 (LOG_ERR, _("unknown tftp user '%s'"), text);
        return 1;
    }
    *uid = sp->pw_uid;
    return 0;
}

int txt2gid (char *text, gid_t *gid)
{
    struct group *sg;
    int igid;

    if ((sg = getgrnam (text)) == NULL) {
        if (sscanf(text,"%d",&igid) == 1)
            sg = getgrgid((gid_t)igid);
    }
    if (sg == NULL) {
        LOG_MSG2 (LOG_ERR, _("unknown tftp group '%s'"), text);
        return 1;
    }
    *gid = sg->gr_gid;
    return 0;
}

int cvtunit (char *text, ulong *unit)
{
    long mul=1, val=0;
    int rtc;

    switch (*(text+strlen(text)-1)) {
      case 'G':
        mul = 1024*1024*1024;
        rtc = sscanf(text,"%ldG",&val);
        break;
      case 'M':
        mul = 1024*1024;
        rtc = sscanf(text,"%ldM",&val);
        break;
      case 'K':
        mul = 1024;
        rtc = sscanf(text,"%ldK",&val);
        break;
      case 'k':
        mul = 1000;
        rtc = sscanf(text,"%ldk",&val);
        break;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        rtc = sscanf(text,"%ld",&val);
        break;
      default:
        return 1;
        break;
    }
    if (rtc != 1)
        return 1;

    *unit = (ulong)mul * (ulong)val;
    return 0;
}

int parseCmr(char *s, dlb_t **bl)
{
    char *t1, *t2, *t3, *ctx1, *ctx2;
    int val1, val2, last=0;

    if ((*bl = initBlistChain()) == NULL)
        return 1;

    for (t1=xstrtok(s,",",&ctx1); t1!=NULL; t1=xstrtok(NULL,",", &ctx1)) {
        if (*t1 == '-') {
            t2 = "1";
            t3 = t1+1;
        }
        else if (*(t1+strlen(t1)-1) == '-') {
            t2 = t1;
            t3 = "2147483647";
        }
        else {
            t2 = xstrtok(t1  ,"-", &ctx2); 
            t3 = xstrtok(NULL,"-", &ctx2);
        }
        if (t3 == NULL) {
            if (sscanf(t2,"%d",&val1) != 1) {
                deleteBlistChain(*bl);
                return 1;
            }
            val2 = val1;
        }
        else {
            if (sscanf(t2,"%d",&val1) != 1) {
                deleteBlistChain(*bl);
                return 1;
            }
            if (sscanf(t3,"%d",&val2) != 1) {
                deleteBlistChain(*bl);
                return 1;
            }
        }
        if ((val1 <= last) || (val2 < val1)) {
            deleteBlistChain(*bl);
            return 1;
        }
        if (addBlist(*bl, val1, val2, NULL)) {
            deleteBlistChain(*bl);
            return 1;
        }
        last = val2;
    }
    return 0;
}

            
void initLocale(char *locale)
{
#ifdef HAVE_LIBINTL_H
    setlocale(LC_ALL, locale);
    if (locale != NULL) {
        bindtextdomain(LOCALEPACK,LOCALEDIR);
        textdomain(LOCALEPACK);
    }
#endif /* HAVE_LIBINTL_H */
}



/*
	compare pattern array with a string
	comparsion may be case insensitive (flag = PAT_NOCASE)
	and / or exact (flag = PAT_EXACT) (no abbreviations permitted)
	result: >= 0 : matching pattern position in pattern struct
	          -1 : no match
	          -2 : ambiguity detected
*/

static int patcomp (patlist_t *pl, char *arg, int flag)
{
    int pc, rt=-1, ac, bc, am;
    char *pat;

    if (flag & PAT_EXACT) {
        for (pc=0; pl->name != NULL; pc++, pl++) {
            if (flag & PAT_NOCASE) {
                if (!strcasecmp(pl->name, arg))
                    return pc;
            }
            else {
                if (!strcmp(pl->name, arg))
                    return pc;
            }
        }
    }
    else {
        for (pc=am=0, bc=-1; pl->name != NULL; pc++, pl++) {
            pat = pl->name;
            for (ac=0; (arg[ac]!='\0') && (pat[ac]!='\0'); ac++) {
                if (flag & PAT_NOCASE) {
                    if ((isupper(pat[ac]) ? tolower(pat[ac]) : pat[ac]) !=
                        (isupper(arg[ac]) ? tolower(arg[ac]) : arg[ac]))
                        break;
                }
                else {
                    if (arg[ac] != pat[ac])
                        break;
                }
            }
            if (!ac)				/* no match at all */
                continue; 
            if (ac == bc)
                am = 1;
            else if (ac > bc) {			/* new best-match */
                bc = ac;
                rt = pc;
                am = 0;
            }
        }
        if (am)					/* unresolved ambiguity ? */
            rt = -2;
    }
    return rt;
}
        
int patstr(patlist_t *pl, char *arg, char *sep, int flag)
{
    char *tok, *ctx, *xarg;
    int pc, rt=0, negate;

    if (sep == NULL)
        sep = ",";

    if ((xarg = Strdup(arg)) == NULL) {
        errNoDynmem;
        return 1;
    }

    for (tok=xstrtok(xarg,sep,&ctx); tok!=NULL; tok=xstrtok(NULL,sep,&ctx)) {
        negate = 0;
        if (flag & PAT_NEGATE) {
            if (*tok == '!') {
                tok++;
                negate=1;
            }
        }
        pc = patcomp(pl, tok, flag);
        switch (pc) {
          case -2:
            Free(xarg);
            return pc;
          case -1:
            continue;
          default:
            rt &= ~(pl+pc)->cval;
            if (negate)
                rt &= ~((pl+pc)->sval);
            else
                rt |=   (pl+pc)->sval;

            if (!(flag & PAT_MULTI)) {
                Free(xarg);
                return rt;
            }
            break;
        }
    }
    Free(xarg);
    return rt;
}

/*
	convert a bitmap to a corresponding parameter list
	this is the reverse operation to patstr()
*/

char *bmap2pat(int val, patlist_t *pl, char *sep)
{
    int sl, size=1;
    char *buf;

    if (sep == NULL)
        sep = ",";

    sl = strlen(sep);

    if ((buf = Malloc(size)) == NULL)
          return NULL;

    *buf = '\0';

    for (; pl->name != NULL; pl++) {
        if (val & pl->sval) {
            size += strlen(pl->name) + sl;
            if ((buf = Realloc(buf, size)) == NULL)
                return NULL;
            strcat (buf, pl->name);
            strcat (buf, sep);
            val &= ~(pl->sval);
        }
    }

    if ((sl = strlen(buf)))
        buf[sl-1] = '\0';

    return buf;
}

/*
	compute the largest power of 2 lower/equal <val>
*/

int powerOf2 (int val)
{
    int i, j;

    for (i=0, j=1; i < (((int)sizeof(int)*8)-1); i++, j *= 2)
        if (j > val)
            return j/2;

    return 0;
}

char *zRealloc(char **p, size_t l)
{
    char *new;

    if (l == 0) {
        free(*p);
        *p = new = NULL;
    }
    else {
        if (((new = realloc(*p, l)) == NULL)) {
            free(*p);
            *p = NULL;
        }
    }
    return new;
}


void zFree(char **p)
{
    free(*p);
    *p = NULL;
}


#ifdef PATTEST

int main (int argc, char **argv)
{
    patlist_t pat[] = {{ "micky", 0x100 , 0},
                       { "minnie",0x200 , 0}, 
                       { "tick",  0x400 , 0}, 
                       { "trick", 0x800 , 0}, 
                       { "track", 0x1000, 0},
                       { NULL   , 0     }};
    char *arg   = "Mick,min,Tick,trick,Tr";
    int  fla[]  = { 0, PAT_NOCASE, 
                    PAT_MULTI, PAT_MULTI|PAT_NOCASE,
                    PAT_EXACT, PAT_EXACT|PAT_NOCASE,
                    PAT_EXACT|PAT_NOCASE|PAT_MULTI};
    patlist_t cfl[] = { { "PAT_NOCASE", PAT_NOCASE , 0},
                        { "PAT_MULTI",  PAT_MULTI  , 0},
                        { "PAT_EXACT",  PAT_EXACT  , 0}, 
                        { NULL       ,  0          , 0}};
    patlist_t *pp;
    int fl, i, rt;

    fl=sizeof(fla)/sizeof(fla[0]);

    printf("\npat:\n");
    for (pp=pat; pp->name != NULL; pp++)
        printf("%8.8x %8.8x (%s)\n", 
               pp->sval,pp->cval,bmap2pat(pp->sval, pat, NULL));
    printf("arg: %s\n", arg);

    for (i=0; i<fl; i++) {
       rt = patstr(pat, Strdup(arg), NULL, fla[i]);
       printf ("flags:  %8.8x (%s)\n", fla[i], bmap2pat(fla[i],cfl,NULL));
       if ((rt == -1) || (rt == -2)) 
           printf ("result: %8d (error)\n", rt);
       else
           printf ("result: %8.8x (%s)\n", rt, bmap2pat(rt,pat,NULL));
    }

    return 0;   
}
#endif /* PATTEST */               


#ifdef TEST
int main (int argc, char **argv)
{
    char *res;
    int rtc; 

    printf ("%d: %s\n", strlen(argv[1]), argv[1]);
    printf ("%d: %s\n", strlen(argv[2]), argv[2]);

    rtc = realPath(argv[1], argv[2], &res);

    printf ("%d: %s\n", rtc, strerror(rtc));
    printf ("%d: %s\n", strlen(argv[1]), argv[1]);
    printf ("%d: %s\n", strlen(argv[2]), argv[2]);
    printf ("%d: %s\n", (res==NULL) ? 0 : strlen(res), res);
    return 0;
}
#endif /* TEST */   
